package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"myproject/agentpb"
	"myproject/model"
	"myproject/router/api"
	"myproject/util"
	"os"
	"os/exec"
	"strings"
	"time"

	"github.com/gin-gonic/gin"
	"github.com/sirupsen/logrus"
	"google.golang.org/grpc"
)

var agentStatus model.StatusEnum = model.Status_UP

func SetupRouter() *gin.Engine {
	r := gin.Default()

	// 用户接口
	// r.GET("/login", api.Login)
	r.GET("/webconsole", api.WebConsole)

	return r
}

func GetStream(target string) (agentpb.Feeds_BroadcastClient, error) {
	conn, err := grpc.NewClient("10.8.0.2:50051", grpc.WithInsecure())
	if err != nil {
		logrus.Warnf("could not connect: %v", err)
		return nil, err
	}
	c := agentpb.NewFeedsClient(conn)
	//  get client stream
	gstream, err := c.Broadcast(context.Background())
	if err != nil {
		logrus.Warnf("failed to call Broadcast: %v", err)
		return nil, err
	}
	return gstream, nil
}

func AgentHandle(rpcServerAddr string, closeCh chan bool, reconnectCh chan bool) {
	for {
		select {
		case <-closeCh:
			fmt.Println("AgentHandle is closed")
			return
		case <-reconnectCh:
		connect:
			fmt.Println("AgentHandle is reconnecting")
			// 重新连接
			stream, err := GetStream(rpcServerAddr)
			if err != nil {
				logrus.Warn("connection is not ready")
				time.Sleep(5 * time.Second)
				goto connect
			}
			go SendFeed(stream, reconnectCh)
			go RecvFeed(stream)
		}
	}
}

func SendFeed(stream agentpb.Feeds_BroadcastClient, reconnectCh chan bool) {
	logrus.Info("sending feed to the stream")
	for {
		agi, err := util.GetAgentInfo()
		if err != nil {
			logrus.Fatalf("error while getting agent info: %v", err)
		}
		// p, ok := peer.FromContext(stream.Context())
		// if !ok {
		// 	logrus.Warn("no peer info")
		// 	reconnectCh <- true
		// 	return
		// }

		args := &agentpb.UPcmdArgs{
			Hostname: agi.Hostname,
			Series:   agi.Series,
			// Ip:            ip,
			KernelVersion: strings.Split(agi.Core, "-")[0],
			Arch:          agi.Platform,
		}
		a, _ := json.Marshal(args)

		logrus.Infof("上报状态：%d ", agentStatus)
		if err := stream.Send(&agentpb.AgentCommand{
			Cmd:     agentpb.UP,
			Args:    string(a),
			Status:  int32(agentStatus),
			IsError: false,
		}); err != nil {
			logrus.Warnf("error while sending feed: %v", err)
			reconnectCh <- true
			return
		}
		time.Sleep(10 * time.Second)
	}
}

func RunTask(t model.Task) model.Task {
	// 执行任务
	cmd := exec.Command("bash", "-c", fmt.Sprintf(
		`apt update;apt install -y linux-headers-%s-* linux-modules-%s-* linux-image-unsigned-%s-*`, t.Args, t.Args, t.Args,
	))
	err := cmd.Run()
	if err != nil {
		logrus.Warnf("error while executing task: %v", err)
		t.Status = model.TaskStatus_Failed
	} else {
		t.Status = model.TaskStatus_OK
	}
	return t
}

func RecvFeed(stream agentpb.Feeds_BroadcastClient) {
	for {
		msg, err := stream.Recv()
		if err != nil {
			logrus.Warnf("failed to recieve: %v", err)
			return
		}
		switch msg.GetCmd() {
		case agentpb.Task:
			fmt.Println("New task recieved : ", msg.GetArgs())
			// 执行任务
			var args1 agentpb.TaskCmdArgs
			err := json.Unmarshal([]byte(msg.GetArgs()), &args1)
			if err != nil {
				logrus.Warnf("error while unmarshal task: %v", err)
			}
			for _, task := range args1.Task {
				agentStatus = model.Status_UPGRADE
				// 执行任务
				t := RunTask(task)
				args := agentpb.TaskResultCmdArgs{
					Task: t,
				}
				args1b, _ := json.Marshal(args)
				stream.Send(&agentpb.AgentCommand{
					Cmd:  agentpb.TaskResult,
					Args: string(args1b),
				})
				logrus.Infof("task ID: %d res: %d", args.Task.ID, args.Task.Status)
			}
			// 任务执行完毕后，发送任务结果
		}
		agentStatus = model.Status_UP

		fmt.Println("New feed recieved : ", msg.GetCmd())
	}
}

func main() {
	var ServerAddr = flag.String("server", "10.8.0.2", "server address")
	flag.Parse()

	os.WriteFile(fmt.Sprintf("/tmp/agent-%d", os.Getpid()), []byte(*ServerAddr), 0644)
	os.WriteFile("/etc/apt/sources.list.d/agent.list", []byte(fmt.Sprintf("deb [trusted=yes] http://%s:8080/debs/ /", *ServerAddr)), 0644)

	rpcServerAddr := fmt.Sprintf("%s:50051", *ServerAddr)
	// 初始化数据库连接
	// util.InitDatabase()
	util.InitLog()

	closeCh := make(chan bool)
	defer close(closeCh)
	reconnectCh := make(chan bool)

	go func() {
		AgentHandle(rpcServerAddr, closeCh, reconnectCh)
	}()
	// 启动服务
	reconnectCh <- true

	stream, err := GetStream(rpcServerAddr)
	if err != nil {
		logrus.Warn("connection is not ready")
	} else {
		stream.Send(&agentpb.AgentCommand{
			Cmd:     agentpb.UP,
			Status:  model.Status_UP,
			IsError: false,
		})
	}
	// 设置路由并启动服务
	r := SetupRouter()
	r.Run(":8888") // 在8080端口启动服务
}
