package api

import (
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"os/exec"
	"time"

	"github.com/creack/pty"
	"github.com/gin-gonic/gin"
	"github.com/gorilla/websocket"
	"github.com/sirupsen/logrus"
)

type WebConsoleInPut struct {
	Host string `json:"host" binding:"required"`
}

// message type enum input | resize
type WebConsoleMessageType string

const (
	WebConsoleInput  WebConsoleMessageType = "input"
	WebConsoleResize WebConsoleMessageType = "resize"
)

// message struct input | resize
type WebConsoleMessage struct {
	Type    WebConsoleMessageType `json:"msg_type"`
	Data    string                `json:"data"`
	Columns int                   `json:"cols"`
	Rows    int                   `json:"rows"`
}

func WebConsole(c *gin.Context) {
	upGrader := websocket.Upgrader{
		CheckOrigin: func(r *http.Request) bool {
			return true
		},
		ReadBufferSize:  1024,
		WriteBufferSize: 1024,
	}

	ws, err := upGrader.Upgrade(c.Writer, c.Request, nil)
	if err != nil {
		logrus.Error(err)
	}

	defer func() {
		closeSocketErr := ws.Close()
		if closeSocketErr != nil {
			logrus.Error(err)
		}
	}()

	command := exec.Command("bash")
	command.Env = append(command.Env, "TERM=xterm-256color")

	// Start the command with a pty.
	ptmx, err := pty.Start(command)
	if err != nil {
		logrus.Error(err)
	}
	exitCh := make(chan error)

	defer ptmx.Close()
	go func() {
		for {
			_, msg, err := ws.ReadMessage()
			if err != nil {
				logrus.Error(err)
				exitCh <- nil
			}
			msgObj := &WebConsoleMessage{}
			if err = json.Unmarshal(msg, msgObj); err != nil {
				return
			}
			switch msgObj.Type {
			case WebConsoleInput:
				ptmx.Write([]byte(msgObj.Data))
			case WebConsoleResize:
				cols := msgObj.Columns
				rows := msgObj.Rows
				if cols == 0 || rows == 0 {
					cols = 80
					rows = 24
				}
				pty.Setsize(ptmx, &pty.Winsize{
					Rows: uint16(rows),
					Cols: uint16(cols),
				})
			case "ping":
				ws.SetReadDeadline(time.Now().Add(10 * time.Second))
			}
		}
	}()
	go func() {
		for {
			data := make([]byte, 1024)
			n, err := ptmx.Read(data)
			if err != nil {
				logrus.Error(err)
				exitCh <- nil
			}
			ws.WriteMessage(websocket.TextMessage, data[:n])
		}
	}()
	<-exitCh
	logrus.Warn(command.ProcessState)
}

type ServerSideWebConsoleInPut struct {
	// url query param host
	Host string `form:"host" binding:"required"`
}

func ServerSideWebConsole(c *gin.Context) {
	upGrader := websocket.Upgrader{
		CheckOrigin: func(r *http.Request) bool {
			return true
		},
		ReadBufferSize:  1024,
		WriteBufferSize: 1024,
	}
	ws, err := upGrader.Upgrade(c.Writer, c.Request, nil)
	if err != nil {
		logrus.Error(err)
	}
	ws.SetReadDeadline(time.Now().Add(10 * time.Second))
	defer func() {
		closeSocketErr := ws.Close()
		if closeSocketErr != nil {
			logrus.Error(err)
		}
	}()
	input := ServerSideWebConsoleInPut{}
	if err = c.ShouldBindQuery(&input); err != nil {
		logrus.Error(err)
	}
	logrus.Info(input)
	u := url.URL{Scheme: "ws", Host: fmt.Sprintf("%s:8888", input.Host), Path: "/webconsole"}
	wsclient, _, err := websocket.DefaultDialer.Dial(u.String(), nil)
	if err != nil {
		logrus.Error(err)
		return
	}
	defer wsclient.Close()
	exitCh := make(chan error)
	// server to broswer
	go func() {
		for {
			_, err := io.Copy(ws.NetConn(), wsclient.NetConn())
			if err != nil {
				logrus.Error(err)
				exitCh <- nil
				break
			}
		}
	}()

	// server to agent
	go func() {
		for {
			// _, err := io.Copy(wsclient.NetConn(), ws.NetConn())
			_, msg, err := ws.ReadMessage()
			if err != nil {
				logrus.Error(err)
				exitCh <- nil
				break
			}
			if err = wsclient.WriteMessage(websocket.TextMessage, msg); err != nil {
				logrus.Error(err)
				exitCh <- nil
				break
			}
			ws.SetReadDeadline(time.Now().Add(10 * time.Second))
		}
	}()

	<-exitCh
}
